/****************************************************************************************************
 * DICTIONARY APP SUPPORT
 */
/*
 Keys starting with _ are generated.  User keys that start with _ may cause undefined behaviour
 or be overwritten.

 ddb {
 	config {
		key {
			ckey {
				enabled = bool
				value = <value>
				layout {
					type = *label | input | checkbox | button
					class = str
					order = int
					label = str
					readonly = bool
					dest = div id
				} | layout = falase
			}
		}
		_akey[]
	}
	log {
		key {
			lkey {
				enabled = bool
				layout {
					dest = div id
					label = str
					height = 600px
					interpolation = bool
					style: bar | *line
					bar {
						align: center | left | right
						width: int
					}
					shaded {
						bottom = bool
					}
					points {
						enabled = bool
						size = int
						style = *circle | square
					}
					data {
						max: num
						min: num
						minorLabel: bool
						title: str
					}
				}
				_id
				_group
			}
		}
	}
}
*/

var DictionaryApp = function (dictionaryname, options) {

	if (options === undefined) {
		options = {
			configid: 'phidgetextra_config',
			logid: 'phidgetextra_log'
		};
	} else {
		if (options.configid === undefined)
			options.configid = 'phidgetextra_config';
		if (options.logid === undefined)
			options.logid = 'phidgetextra_log';
	}
	this.options = options;

	this.dname = dictionaryname;
	this.class = 'phidget phidget_da phidget_da_' + this.dname;

	this.mkHTML = function (cfg) {

		switch (cfg.layout.type) {
		default:
		case 'label':
			return (this.mkLabelHTML(cfg));
		case 'checkbox':
			return (this.mkCheckBoxHTML(cfg));
		case 'input':
			return (this.mkInputHTML(cfg));
		case 'button':
			return (this.mkButtonHTML(cfg));
		}
	}

	this.mkCheckBoxHTML = function (cfg) {
		var html = [];

		html.push('<input id="' + cfg._id + '" type="checkbox"');
		if (cfg.layout.readonly)
			html.push(' disabled="disabled"');
		if (cfg.value)
			html.push(' checked');
		if (cfg.layout.class)
			html.push(' class="' + cfg.layout.class + '"');
		html.push('>');

		html = html.join(' ');
		return (html);
	};

	this.mkLabelHTML = function (cfg) {
		var html = [];

		html.push('<label id="' + cfg._id + '"');
		if (cfg.layout.class)
			html.push(' class="' + cfg.layout.class + '"');
		html.push('>');

		if (cfg.value !== undefined)
			html.push(cfg.value);

		html.push('</label>');

		return (html.join(' '));
	};

	this.mkInputHTML = function (cfg) {
		var html = [];

		html.push('<input id="' + cfg._id + '"');
		if (cfg.layout.class)
			html.push(' class="' + cfg.layout.class + '"');
		if (cfg.value !== undefined)
			html.push(' value="' + cfg.value + '"');
		html.push('>');

		return (html.join(' '));
	};

	this.mkButtonHTML = function (cfg) {
		var html = [];

		var label = cfg._key;
		if (cfg.layout.label)
			label = cfg.layout.label;

		html.push('<input type="button" id="' + cfg._id + '"');
		if (cfg.layout.class)
			html.push(' class="' + cfg.layout.class + '"');
		html.push(' value="' + label + '"');
		html.push('>');

		return (html.join(' '));
	};
};

/*
 * Expects to be bound to the log object.
 */
function getLogData() {

	var log = this;

	var options = {
		startid: log._nextid !== undefined ? log._nextid : 0
	}

	log._parent.getLog(log._key, options).then(function (data) {
		if (log.layout.graph)
			var lc = log._parent.log[log.layout.graph];
		else
			var lc = log._parent.log[log._key];

		if (data.data.length === 0)
			return;

		for (var r in data.data) {
			r = data.data[r];
			r.y = r.val;
			r.x = new Date(r.time);
			r.group = log._group;
			log._nextid = r.id + 1;
		}
		lc.dataset.add(data.data);

		if (log._first === true) {
			lc.graph2d.setWindow(data.data[0].x,  data.data[data.data.length - 1].x);
			log._first = false;
		}

		/* off by default */
		if (lc.moving) {
			var now = vis.moment();
			var range = lc.graph2d.getWindow();
			var interval = range.end - range.start;
			if (now > range.end)
				lc.graph2d.setWindow(now - 0.5 * interval, now + 0.5 * interval);
		}
	});
}

DictionaryApp.prototype.setupDDB = function () {

	/* CONFIG */
	if (this.ddb.config === undefined)
		this.ddb.config = { key: {} };
	if (this.ddb.config.key === undefined)
		this.ddb.config.key = {};

	var config = this.ddb.config;
	var key = [];
	for (var k in config.key) {
		var cfg = config.key[k];
		if (cfg.enabled === false)
			continue;

		cfg._id = this.dname + '_' + k;
		cfg._key = k;

		if (cfg.layout === undefined)
			cfg.layout = { type: 'label' };
		if (cfg.layout.type === undefined)
			cfg.layout.type = 'label';

		key.push(cfg);
	}
	key.sort(function (a, b) {
		if (a.layout.order !== undefined) {
			if (b.layout.order == undefined)
				return (-1);
			return (a.layout.order - b.layout.order);
		}
		if (b.layout.order != undefined) {
			if (a.layout.order == undefined)
				return (1);
			return (a.layout.order - b.layout.order);
		}
	});
	config._akey = key;

	/* LOG */

	if (this.ddb.log === undefined)
		this.ddb.log = { key: {} };
	if (this.ddb.log.key === undefined)
		this.ddb.log.key = {};

	var logs = this.ddb.log;

	var group = 0;
	this.log = {};
	for (var k in logs.key) {
		var log = logs.key[k];
		if (log.enabled === false)
			continue;

		log._parent = this;
		log._key = k;
		log._id = k + '_loggraph';
		log._group = group++;
		log._first = true;

		if (log.layout === undefined)
			log.layout = {};
		if (log.layout.interpolation === undefined)
			log.layout.interpolation = false;
		if (log.layout.height === undefined)
			log.layout.height = '400px';
		if (log.layout.graph === undefined)
			log.layout.graph = log._key
		if (log.layout.label === undefined)
			log.layout.label = log._key
		if (log.layout.shaded !== undefined) {
			if (log.layout.shaded.bottom === undefined || log.layout.shaded.bottom === true)
				log.layout.shaded.orientation = 'bottom';
			else
				log.layout.shaded.orientation = 'top';
		}

		var lc = this.log[log.layout.graph];
		if (lc === undefined) {
			lc = this.log[log.layout.graph] = {
				key: log.layout.graph,
				dataset: new vis.DataSet({queue: { delay: 500, max: 250 }}),
				options: {},
				members: [log],
				groups: new vis.DataSet(),
			};
		} else {
			lc.members.push(log);
		}

		/*
		 * Only bother with the graph layout options if we are creating a graph.
		 */
		if (log.layout.graph === log._key) {
			lc.options.legend = true;
			lc.options.interpolation = log.layout.interpolation;
			lc.options.height = log.layout.height;

			if (log.layout.data !== undefined) {
				lc.options.dataAxis = { left: { range: {} } };
				if (log.layout.data.min !== undefined)
					lc.options.dataAxis.left.range.min = log.layout.data.min;
				if (log.layout.data.max !== undefined)
					lc.options.dataAxis.left.range.max = log.layout.data.max;
				if (log.layout.data.title !== undefined)
					lc.options.dataAxis.left.title = { text: log.layout.data.title };
				if (log.layout.data.minorLabel === false)
					lc.options.dataAxis.showMinorLabels = false;
			}

			if (log.layout.points !== undefined) {
				if (log.layout.points === false) {
					lc.options.drawPoints = { enabled: false };
				} else {
					if (log.layout.points.enabled === false)
						lc.options.drawPoints = { enabled: false };
					else
						lc.options.drawPoints = { enabled: true };

					if (log.layout.points.size !== undefined)
						lc.options.drawPoints.size = log.layout.points.size;
					if (log.layout.points.style !== undefined) {
						if (log.layout.points.style === 'circle')
							lc.options.drawPoints.style = 'circle'
						else if (log.layout.points.style === 'square')
							lc.options.drawPoints.style = 'square'
					}
				}
			} else {
				lc.options.drawPoints = { enabled: true, style: 'circle' };
			}
		}

		/*
		 * The main is used to get layout options etc.
		 * If they graph name is the key name, use that; otherwise, use the last
		 * log we find.
		 */
		if (lc.key == log._key)
			lc.main = log;
		if (lc.main === undefined)
			lc.main = log;
	}

	for (var lc in this.log) {
		lc = this.log[lc];
		lc.moving = false;

		for (var log in lc.members) {
			log = lc.members[log]; 
			var group = {
				id: log._group,
				content: log.layout.label,
				options: {
					interpolation: log.layout.interpolation
				}
			};
			if (log.layout.shaded !== undefined && log.layout.shaded.orientation !== undefined)
				group.options.shaded = { orientation: log.layout.shaded.orientation };
			lc.groups.add(group);
		}
	}
}

DictionaryApp.prototype.start = function () {

	var self = this;

	return (new Promise(function (resolve, reject) {
		self.getDictionary().then(function (ddb) {
			self.ddb = ddb;

			self.setupDDB();

			self.addHTMLComponents();
			self.addLogComponents();

			self.dict = new jPhidgets.Dictionary();
			self.dict.setDeviceLabel(self.ddb.label);
			self.dict.setDeviceSerialNumber(self.ddb.sn);
			self.dict.onAttach = self.onDictionaryAttach.bind(self);
			self.dict.onDetach = self.onDictionaryDetach.bind(self);
			self.dict.onAdd = self.onDictionaryAdd.bind(self);
			self.dict.onUpdate = self.onDictionaryUpdate.bind(self);
			self.dict.onRemove = self.onDictionaryRemove.bind(self);
			self.dict.open().then(function () {
				resolve();
			}).catch(function (err) {
				reject(err);
			});

		}).catch(function (err) {
			reject(err);
		});
	}));
};

DictionaryApp.prototype.getDictionary = function () {

	var self = this;

	return (new Promise(function (resolve, reject) {
		var jqxhr = $.get('/api/v1/dictionary?name=' + self.dname + '&action=get');
		jqxhr.fail(function (jqXHR, textStatus, errorThrown) {
			console.log('getDictionary() failed:' + textStatus);
			console.log(errorThrown);
			reject(errorThrown);
		});
		jqxhr.done(function (data) {
			resolve(data.dictionary);
		});
	}));
}

DictionaryApp.prototype.getLog = function (key, query) {

	var self = this;
	var url = '/api/v1/dictionary?name=' + self.dname + '&action=data';

	url += '&key=' + key;

	if (query !== undefined) {
		for (var q in query)
			url += '&' + q + '=' + query[q];
	}

	//console.log('url:' + url);

	return (new Promise(function (resolve, reject) {
		var jqxhr = $.get(url);
		jqxhr.fail(function (jqXHR, textStatus, errorThrown) {
			console.log('getLog() failed:' + textStatus);
			console.log(errorThrown);
			reject(errorThrown);
		});
		jqxhr.done(function (data) {
			resolve(data);
		});
	}));
}

DictionaryApp.prototype.onDictionaryAttach = function () {

	var self = this;

	console.log('Dictionary Attached');
	if (this.userOnDictionaryAttach)
		this.userOnDictionaryAttach(this);

};

DictionaryApp.prototype.onDictionaryDetach = function () {

	console.log('Dictionary Detached');
	if (this.userDictionaryDetach)
		this.userDictionaryDetach(this);
};

DictionaryApp.prototype.onDictionaryAdd = function (key, val) {

	if (this.options.onAdd)
		this.options.onAdd(key, val);
}

DictionaryApp.prototype.onDictionaryRemove = function (key) {

	if (this.options.onRemove)
		this.options.onRemove(key);
}

DictionaryApp.prototype.onDictionaryUpdate = function (key, val) {

	var cfg = this.ddb.config.key[key];
	if (cfg == undefined)
		return;

	if (cfg.layout === false)
		return;

	if (this.options.onUpdate) {
		if (this.options.onUpdate(key, val) === true)
			return;
	}

	switch (cfg.layout.type) {
	case 'checkbox':
		$('#' + cfg._id).prop('checked', val === '0' ? false : true);
		break;
	case 'label':
		$('#' + cfg._id).text(val);
		break;
	case 'input':
		$('#' + cfg._id).val(val);
		break;
	default:
		console.log('unexpected update for ' + key);
	}
}

DictionaryApp.prototype.onAppKey = function (key, val) {

	key = key.substring(this.appname.length + 1);
	if (key.endsWith('.layout')) {
		this.handleLayout(key, val, true);
		return;
	}

	if (this.cb === undefined || this.cb(key, val) === false)
		this.defaultValueHandler(key, val);
};

DictionaryApp.prototype.mkHTMLComponentLabel = function (cfg) {

	var label = cfg._key;
	if (cfg.layout.label)
		label = cfg.layout.label;

	return ('<label id="' + cfg._id + '_label">' + label + '</label>');
}

DictionaryApp.prototype.mkHTMLComponent = function (cfg) {

	if (cfg.layout.class)
		cfg.layout.class += this.class;
	else
		cfg.layout.class = this.class;

	var ele = $('#' + cfg._id);
	if (ele.length) {
		console.log('dictionary key id: "' + cfg._id + '" already defined');
		return;
	}

	return (this.mkHTML(cfg));
}

DictionaryApp.prototype.addHTMLComponentHandler = function (cfg) {

	if (cfg.layout.events === false)
		return;

	switch (cfg.layout.type) {
	case 'checkbox':
		$(document).on('change', '#' + cfg._id, { da: this, cfg: cfg }, this.checkBoxToggle);
		break;
	case 'input':
		$(document).on('change', '#' + cfg._id, { da: this, cfg: cfg }, this.inputChange);
		break;
	case 'button':
		$(document).on('click', '#' + cfg._id, { da: this, cfg: cfg }, this.buttonClick);
		break;
	}
};

DictionaryApp.prototype.addHTMLComponents = function () {

	this.udiv = $('#' + this.options.configid);
	if (!this.udiv.length) {
		console.log('missing target div:' + this.options.configid);
		return;
	}

	/*
	 * Add config entries that do not specify their own location id.
	 */
	var html = [];
	html.push('<table class="' + this.class + '">');
	for (var key in this.ddb.config._akey) {
		var cfg = this.ddb.config._akey[key];

		if (cfg.layout === false || cfg.layout.dest !== undefined)
			continue;

		html.push('<tr class="' + this.class + '">');
		if (cfg.layout.type !== 'button') {
			html.push('<td class="' + this.class + '">');
			html.push(this.mkHTMLComponentLabel(cfg));
			html.push('</td>');
			html.push('<td class="' + this.class + '">');
			html.push(this.mkHTMLComponent(cfg));
			html.push('</td>');
		} else {
			html.push('<td colspan="2" class="' + this.class + '">');
			html.push(this.mkHTMLComponent(cfg));
			html.push('</td>');
		}
		html.push('</tr>');
	}
	html.push('</table>');
	this.udiv.append(html.join(' '));

	/*
	 * Add config entries that specify their own location id.
	 */
	for (var key in this.ddb.config._akey) {
		var cfg = this.ddb.config._akey[key];

		if (cfg.layout === false || cfg.layout.dest === undefined)
			continue;
		var udiv = $('#' + cfg.layout.dest);
		if (udiv.length === 0) {
			console.log(cfg._key + ' destination id (' + cfg.layout.dest + ') not found');
			continue;
		}
		if (cfg.layout.label !== undefined)
			udiv.append(this.mkHTMLComponentLabel(cfg));
		udiv.append(this.mkHTMLComponent(cfg));
	}

	/*
	 * Add Event Handlers
	 */
	for (var key in this.ddb.config.key) {
		var cfg = this.ddb.config.key[key];
		if (cfg.layout === false)
			continue;
		this.addHTMLComponentHandler(cfg);
	}
};

DictionaryApp.prototype.addLogComponents = function () {

	this.logdiv = $('#' + this.options.logid);
	if (!this.logdiv.length) {
		console.log('missing log graph div:' + this.options.logid);
		return;
	}

	for (var lc in this.log) {
		lc = this.log[lc];

		var div = '<div id="' + lc.main._id + '" class="phidget_da_graph"></div>';
		var udiv = this.logdiv;

		if (lc.main.layout.dest !== undefined) {
			udiv = $('#' + lc.main.layout.dest);
			if (udiv.length === 0) {
				console.log(lc.key + ' destination id (' + lc.main.layout.dest + ') not found');
				continue;
			}
		}
		udiv.append(div);
		lc.graph2d = new vis.Graph2d($('#' + lc.main._id)[0], lc.dataset, lc.groups, lc.options);

		for (var log in lc.members) {
			log = lc.members[log];
			getLogData.call(log);
			log._intervalId = setInterval(getLogData.bind(log), 5000);
		}
	}
}

DictionaryApp.prototype.defaultValueHandler = function (key, val) {

	var cfg = this.ddb.config.key[key];
	if (cfg === undefined) {
		console.log('key: ' + key + ' has not been seen before; ignoring');
		return;
	}

	switch (cfg.layout.type) {
	case 'checkbox':
		if (val == 0)
			$('#' + cfg._id).prop('checked', false);
		else
			$('#' + cfg._id).prop('checked', tru);
		break;
	case 'label':
		$('#' + cfg._id).text(val);
		break;
	case 'input':
		$('#' + cfg._id).text(val);
		break;
	default:
		return (false);
	}

	return (true);
}

DictionaryApp.prototype.add = function (key, val) {

	return (this.dict.add(key, val));
}

DictionaryApp.prototype.set = function (key, val) {

	return (this.dict.set(key, val));
}

DictionaryApp.prototype.update = function (key, val) {

	return (this.dict.update(key, val));
}

DictionaryApp.prototype.get = function (key, def) {

	return (this.dict.get(key, def));
}

DictionaryApp.prototype.remove = function (key) {

	if (this.ddb.config.key[key] !== undefined)
		throw (new Error('Attempt to remove config key:' + key));

	return (this.dict.remove(key));
}

DictionaryApp.prototype.checkBoxToggle = function (ev) {

	ev.data.da.dict.set(ev.data.cfg._key, this.checked);
};

DictionaryApp.prototype.inputChange = function (ev) {

	ev.data.da.dict.set(ev.data.cfg._key, $(this).val());
};

DictionaryApp.prototype.buttonClick = function (ev) {

	ev.data.da.dict.set('__clicked__', ev.data.cfg._key);
};

/****************************************************************************************************
 * DICTIONARY UTILS
 */

var DictionaryUtils = function (dict) {

	this.dict = dict;

	this.dict.onUpdate = function (key, val) {

		for (var i in this.regs) {
			var reg = this.regs[i];
			if (reg.key instanceof RegExp) {
				if (reg.key.test(key))
					reg.cb(key, val);
			} else {
				if (reg.key == key)
					reg.cb(key, val);
			}
		}
	}.bind(this);

	this.regs = [];
};

DictionaryUtils.prototype.on = function (key, cb) {

	this.regs.push({ key: key, cb: cb });
};

DictionaryUtils.prototype.off = function (key) {

	for (var i in this.regs) {
		if (this.regs[i].key === key) {
			this.regs.splice(i, 1);
			return;
		}
	}
};

DictionaryUtils.prototype.toBool = function toBool(val) {

	if (val === undefined || val === null)
		return (false);

	switch (val.toLowerCase().trim()) {
		case 'false':
		case 'no':
		case '0':
			return (false);
		default:
			return (true);
	}
};

DictionaryUtils.prototype.mkTemperature = function (temp) {

	return (parseFloat(temp).toFixed(2) + 'C');
};

DictionaryUtils.prototype.mkHumidity = function (hum) {

	return (parseFloat(hum).toFixed(2) + '%');
};

DictionaryUtils.prototype.mkLux = function (lux) {

	return (parseFloat(lux).toFixed(2) + 'lx');
};

DictionaryUtils.prototype.mkDB = function (dB) {

	return (parseFloat(dB).toFixed(2) + 'dB');
};

DictionaryUtils.prototype.getAlias = function (key) {

	var parts = key.split(':');
	return (parts[0]);
};

DictionaryUtils.prototype.getAliasTarget = function (key) {

	var parts = key.split(':');
	return (parts[1]);
};

DictionaryUtils.prototype.getAliasID = function (key) {

	var parts = key.split(':');
	if (parts.length < 3)
		return ('#' + parts[0]);
	return ('#' + parts[0] + '_' + parts[2]);
};

DictionaryUtils.prototype.fireAll = function (sort) {

	if (sort)
		var values = this.values.sort(sort);
	else
		var values = this.values;

	for (var off in values)
		this.dict.onUpdate(values[off].key, values[off].val);
};

DictionaryUtils.prototype.load = function () {

	this.values = [];
	this.keys = [];
	var self = this;

	function getKeys(dict, startKey) {

		return (new Promise(function (resolve, reject) {

			dict.scan(startKey).then(function (keys) {
				resolve(keys);
			}).catch(function (err) {
				reject(err);
			});
		}).then(function (keys) {
			if (keys.length == 0)
				return;
			self.keys = self.keys.concat(keys);
			return (getKeys(dict, self.keys[self.keys.length - 1]));
		}));
	}

	function getkey(dict, cfg, key) {

		return (new Promise(function (resolve, reject) {
			dict.get(key).then(function (val) {
				cfg.values.push({ key: key, val: val});
				resolve(cfg);
			}).catch(function (err) {
				reject(err);
			});
		}));
	}

	return (new Promise(function (resolve, reject) {

		getKeys(self.dict, '').then(function () {
			var promises = [];
			for (var k of self.keys)
				promises.push(getkey(self.dict, self, k));

			Promise.all(promises).then(function () {
				resolve(self);
			}).catch(function (err) {
				reject(err);
			});
		});
	}));
};

if (typeof module !== 'undefined' && module.exports)
	module.exports = DictionaryUtils;
